% ***************************************************************
% author:   Wei Hu                                             **
% date:     04/12/2013                                         **
% modified: 01/13/2015                                         **
% purpose:  perform NLGBT denoising                            **
% input:    y: noisy image                                     **
%           x: original image for computing PSNR               **
%           nSig: noise variance                               **
% output:   im_out: denoised image                             **
%           psnr: the PSNR of the denoised image               **
% ***************************************************************

function [im_out,psnr] = NLGBT(y,x,nSig)

im_out = y;
[h, w] = size(y);
par.nSig = nSig;

% parameters
if nSig <= 15
    par.win       =   5;      % window size
    par.nblk      =   20;     % number of K-nearest-neighbors 
    par.lamada    =   0.63;   % parameter for noise variance update
    par.w         =   0.23;   % parameter for iterative regularization
    par.K         =   6;      % number of iterations
elseif nSig <= 30
    par.win       =   5;
    par.nblk      =   20;
    par.lamada    =   0.65;
    par.w         =   0.23;
    par.K         =   11;
elseif nSig <= 50
    par.win       =   8;
    par.nblk      =   75;
    par.lamada    =   0.67;
    par.w         =   0.23;    
    par.K         =   12;
else
    par.win       =   9;
    par.nblk      =   90;
    par.lamada    =   0.64;
    par.w         =   0.23;    
    par.K         =   14;    
end
par.step = min(6, par.win-1);
lambda = par.w;

for iter = 1 : par.K

    % step 1: iterative regularization & noise variance update
    im_out  =   im_out + lambda*(y - im_out);
    
    dif     =   im_out - y;
    vd      =   nSig^2 - (mean(mean(dif.^2)));
        
    if iter==1
        par.nSig  = sqrt(abs(vd));            
    else
        par.nSig  = sqrt(abs(vd))*par.lamada;
    end    
    
    % step 2: patch clustering
    if (mod(iter,6)==0) || (iter==1)
        blk_arr   =   Block_matching( im_out, par);
    end

    X  =  Im2Patch( im_out, par );    
    Ys =  zeros( size(X) );   
    W  =  zeros( size(X) );
    L = size(blk_arr,2);
    
    for  i  =  1 : L

        B = double(X(:, blk_arr(:, i)));  % group of similar patches
        
        %% step 3: NLGBT denoising
        % step 3.1: compute the average patch for a group of similar patches       
        mP = mean(B,2);
        mblk = reshape(mP,par.win,par.win);
        
        % step 3.2: compute the weighted GBT basis functions based on the average patch 
        [V] = WGBT(mblk,par.nSig);

        % step 3.3: hard-thresholding
        A = V'*B;    % the GBT coefficient matrix
        th = (par.nSig)*sqrt(2*log((par.win)^2*(par.nblk)));
        A(abs(A)<th) = 0;
        
        % step 3.4: reconstruction
        B_hat = round(V*A);       
        r = rank(A);
        if r==size(B,1)
            wei = 1/size(B,1);
        else
            wei = (size(B,1)-r)/size(B,1);
        end
        W(:, blk_arr(:,i)) = wei*ones( size(B_hat) );  
        Ys(:, blk_arr(:,i)) = (B_hat)*wei;
    end

    % step 4: image update
    im_out   =  zeros(h,w);
    im_wei   =  zeros(h,w);
    k        =   0;
    b        =   par.win;
    N        =   h-b+1;
    M        =   w-b+1;
    r        =   [1:N];
    c        =   [1:M]; 
    for i  = 1:b
        for j  = 1:b
            k    =  k+1;
            im_out(r-1+i,c-1+j)  =  im_out(r-1+i,c-1+j) + reshape( Ys(k,:)', [N M]);
            im_wei(r-1+i,c-1+j)  =  im_wei(r-1+i,c-1+j) + reshape( W(k,:)', [N M]);
        end
    end
    im_out  =  im_out./(im_wei+eps);
    PSNR(iter) = csnr( im_out, x, 0, 0 );
    
end 

psnr = max(PSNR);

end